#include <linux/slab.h>
#include <linux/err.h>

#include "debug.h"

#include "scsi_codes.h"
#include "scsidump_ctlr.h"
#include "scsidump.h"
#include "filter.h"

extern struct scsi_dmp_ctlr dump_ctlr;

int is_rule_valid(int rule){
	return (rule == RULE_EXC  || rule == RULE_INC);
}


int is_filter_enable(struct scsi_cmd_filter * filter){
	return filter->enable == FILTER_ENABLE;
}

void enable_filter(struct scsi_cmd_filter * filter){
	if(filter == NULL){
		return;
	}

	filter->enable = FILTER_ENABLE;
	return;
}

void disable_filter(struct scsi_cmd_filter * filter){
	if(filter == NULL){
		return;
	}

	kdebug("disable filter, previous state %d\n",filter->enable);
	filter->enable = FILTER_DISABLE;
	return;
}


int  alloc_filter(int rule, char * codes, int codeslen, struct scsi_cmd_filter ** filter){
	int i;
	if(filter == NULL || !is_rule_valid( rule) || (codeslen == 0 && codes != NULL)){
		kdebug("wrong filter rule:%d\n", rule);
		return -EINVAL;
	}

	if(codes != NULL){
		//check codes
		for(i = 0; i < codeslen; i ++){
			if(find_code(codes[i]) == UNKNOWN_CODE){
				kdebug("unknown code 0x%x\n", codes[i]);
				return -EINVAL;
			}
		}
	}

	*filter =  (struct scsi_cmd_filter *) kmalloc(sizeof(scsi_cmd_filter_t), GFP_KERNEL);

	if(*filter == NULL){
		printk("memory allocation error\n");
		return -ENOMEM;
	}

	(*filter)->rule = rule;
	(*filter)->codes = codes;
	(*filter)->codeslen = codeslen;
	(*filter)->enable = FILTER_DISABLE;
	mutex_init( &(*filter)->filter_lock);

	return 0;
	

}
void free_filter(struct scsi_cmd_filter * filter){
	if(filter == NULL){
		return;
	}

	if(filter->codes != NULL){
		kfree(filter->codes);
	}

	kfree(filter);
	return;
}

int modify_filter(struct scsi_cmd_filter * filter, int newrule, char * newcodes, int newcodeslen){
	
	if(filter == NULL  ){
		return -EINVAL;
	}
	if(newcodes == NULL && newcodes != 0){
		return -EINVAL;
	}

	if(!is_rule_valid(newrule)){
		return -EINVAL;
	}

	filter ->rule = newrule;
	if(filter->codes != NULL){
		kfree(filter->codes);
	}

	//copy the codes because we will free scsidump_ioctl 

	filter->codes = (char *)kmalloc (newcodeslen, GFP_KERNEL);
	
	if(filter->codes == NULL){
		printk("out of memory\n");
		return -ENOMEM;
	}

	memcpy(filter->codes, newcodes, newcodeslen);
	
	filter->codeslen = newcodeslen;

	return 0;

}


int do_filter(struct scsi_cmd_filter * filter, char code){
	int i;

	kdebug("filter code 0x%x\n",code);
	if(filter == NULL){
		return -EINVAL;
	}

	if(!is_filter_enable( filter)){
		return -EPERM;
	}

	for(i = 0; i < filter->codeslen; i ++){
		if(filter->codes[i] == code){
			if(filter->rule == RULE_INC){
				return FILTER_PASS;
			}
			else{
				return FILTER_BLOCK;
			}
		}
	}

	//can not find the code in filter

	if(filter->rule == RULE_INC){
		return FILTER_BLOCK;
	}
	else{
		return FILTER_PASS;
	}

}

