
#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/vmalloc.h>
#include <linux/ioctl.h>
#include <asm/uaccess.h>
#include <linux/err.h>
#include <linux/fs.h>

#include "debug.h"
#include "scsi_codes.h"
#include "scsidump_ctlr.h"
#include "scsidump.h"
#include "filter.h"
#include "scsidump_ioctl.h"
extern struct scsi_dmp_ctlr dump_ctlr;

typedef int (*scsidump_ioctl_fn)(struct scsidump_ioctl * data);

int scsidump_ioctl_filter_create(struct scsidump_ioctl * data){
	int rule, codesize;
	char * codes;
	int r;
	struct scsi_cmd_filter * filter;
	if(data == NULL || data->type != SCSIDUMP_IOCTL_TYPE_FILTER ||data->datasize <= 0 || data->data == NULL){
		return -EINVAL;
	}

	//free existing filter
	mutex_lock(&dump_ctlr.filter->filter_lock);
	free_filter(dump_ctlr.filter);
	
	
	rule = (int) data->data[0];
	codesize = data->datasize - 1;
	if(codesize == 0){
		codes = NULL;
	}
	else{
		codes = data->data + 1;
	}

	r = alloc_filter(rule, codes, data->datasize - 1, &filter);
	if(r){
		return r;
	}

	dump_ctlr.filter = filter;
	mutex_unlock(&dump_ctlr.filter->filter_lock);
	return 0;
}

int scsidump_ioctl_filter_enable(struct scsidump_ioctl * data){
	if(data->type != SCSIDUMP_IOCTL_TYPE_FILTER){
		return -EINVAL;
	}
	if(dump_ctlr.filter == NULL){
		return - EPERM;
	}

	mutex_lock(&dump_ctlr.filter->filter_lock);
	enable_filter(dump_ctlr.filter);
	mutex_unlock(&dump_ctlr.filter->filter_lock);
	
	return 0;
}

int scsidump_ioctl_filter_disable(struct scsidump_ioctl * data){
	kinfo("disable filter");
	if(data->type != SCSIDUMP_IOCTL_TYPE_FILTER){
		return -EINVAL;
	}
	if(dump_ctlr.filter == NULL){
		kinfo("filter is NULL");
		return - EPERM;
	}

	mutex_lock(&dump_ctlr.filter->filter_lock);
	disable_filter(dump_ctlr.filter);
	mutex_unlock(&dump_ctlr.filter->filter_lock);
	
	return 0;
}

int scsidump_ioctl_filter_modify(struct scsidump_ioctl * data){
	int rule, codesize;
	char * codes;
	int r;
	if(data == NULL || data->type != SCSIDUMP_IOCTL_TYPE_FILTER ||data->datasize <= 0 || data->data == NULL){
		return -EINVAL;
	}

	if(dump_ctlr.filter == NULL){
		return -EPERM;
	}

	rule = (int) data->data[0];
	codesize = data->datasize - 1;
	if(codesize == 0){
		codes = NULL;
	}
	else{
		codes = data->data + 1;
	}

	kdebug("modify filter in ioctl: rule %d, codes %p, codeslen %d\n", rule, codes, codesize);
	mutex_lock(&dump_ctlr.filter->filter_lock);
	r = modify_filter(dump_ctlr.filter, rule, codes, codesize);
	mutex_unlock(&dump_ctlr.filter->filter_lock);
	if(r){
		return r;
	}

	return 0;
}

int scsidump_ioctl_filter_get(struct scsidump_ioctl * data){
	
	
	if(data == NULL || data->type != SCSIDUMP_IOCTL_TYPE_FILTER 
		||data->datasize  != SCSIDUMP_MAX_CODES_LEN|| data->data == NULL){
		return -EINVAL;
	}

	if(dump_ctlr.filter == NULL){
		return -EPERM;
	}

	kdebug("get filter: rule %d, codeslen %d\n", dump_ctlr.filter->rule, dump_ctlr.filter->codeslen);
	mutex_lock(&dump_ctlr.filter->filter_lock);
	data->data[0] = (char)dump_ctlr.filter->rule;
	data->data[1] = (char) dump_ctlr.filter->enable;
	memcpy(data->data + 1, dump_ctlr.filter->codes, dump_ctlr.filter->codeslen);
	data->datasize = dump_ctlr.filter->codeslen + 2;
	mutex_unlock(&dump_ctlr.filter->filter_lock);
	
	
	return 0;
}

static scsidump_ioctl_fn lookup_ioctl(unsigned int cmd)
{
	static struct {
		int cmd;
		scsidump_ioctl_fn fn;
	} _ioctls[] = {
		{SCSIDUMP_IOCTL_CMD_FILETER_CREATE, scsidump_ioctl_filter_create },	
		{SCSIDUMP_IOCTL_CMD_FILETER_ENABLE,scsidump_ioctl_filter_enable},
		{SCSIDUMP_IOCTL_CMD_FILETER_DISABLE, scsidump_ioctl_filter_disable},
		{SCSIDUMP_IOCTL_CMD_FILETER_MODIFY,scsidump_ioctl_filter_modify},
		{SCSIDUMP_IOCTL_CMD_FILETER_GET,scsidump_ioctl_filter_get}
	};

	return  _ioctls[cmd].fn;
}



static long scsidump_ctl_ioctl(struct file *file, uint command, ulong u){
	struct scsidump_ioctl * userioctl, * kernelioctl;
	scsidump_ioctl_fn fn;
	int r;
	int cmd;

	userioctl = (struct scsidump_ioctl __user *) u;
	if(userioctl == NULL){
		return -EINVAL;
	}
	if(_IOC_TYPE(command) != SCSIDUMP_IOCTL){
		kdebug("wrong command %d\n", command);
		return -ENOTTY;
	}

	cmd = _IOC_NR(command);

	fn = lookup_ioctl(cmd);
	
	kdebug(" command %x\n", cmd);
	if(fn == NULL){

		kdebug("no function for command %x\n", cmd);
		return -EINVAL;
	}

	
	kernelioctl = (struct scsidump_ioctl *)vmalloc(sizeof(struct scsidump_ioctl) + userioctl->datasize);
	if(!kernelioctl){
		return -ENOMEM;
	}
	kdebug("copy from user data size:%d\n",userioctl->datasize );
	if(copy_from_user(kernelioctl, userioctl, sizeof(struct scsidump_ioctl) + userioctl->datasize)){
		vfree(kernelioctl);
		return -EFAULT;
	}

	r = fn(kernelioctl);

	kdebug("fn result %d\n",r);
	if(r){
		return r;
	}

	if(_IOC_DIR(command) & _IOC_READ){
		if(copy_to_user(userioctl, kernelioctl, sizeof(struct scsidump_ioctl) + kernelioctl->datasize)){
			return -EFAULT;
		}
	}

	vfree(kernelioctl);

	return r;

}


static const struct file_operations _ctl_fops = {
	.unlocked_ioctl	 = scsidump_ctl_ioctl,
	.compat_ioctl = NULL,
	.owner	 = THIS_MODULE,
};

static struct miscdevice _scsidump_misc = {
	.minor 		= MISC_DYNAMIC_MINOR,
	.name  		= SCSIDUMP_NAME,
	.fops  		= &_ctl_fops
};

/*
 * Create misc character device and link to DM_DIR/control.
 */
int __init scsidump_interface_init(void)
{
	int r;

	r = misc_register(&_scsidump_misc);
	if (r) {
		kdebug("misc_register failed for control device, error:%d", r);
		return r;
	}
	return 0;
}

void scsidump_interface_exit(void)
{
	int r;
	if ((r = misc_deregister(&_scsidump_misc)) < 0)
		kdebug("misc_deregister failed for control device, error %d",r);
}

